@@ -26,6 +26,11 @@ from linkgrammar import (Sentence, Linkage, ParseOptions, Link, Dictionary,
                          Clinkgrammar as clg)
 
 print(clg.linkgrammar_get_configuration())
+NO_SQLITE_ERROR = ''
+# USE_SQLITE is currently not used in the MSVC build.
+if re.search(r'_MSC_FULL_VER', clg.linkgrammar_get_configuration()) and \
+   not re.search(r'USE_SQLITE', clg.linkgrammar_get_configuration()):
+    NO_SQLITE_ERROR = 'Library is not configures with SQLite support'
 
 # Show the location and version of the bindings modules
 for imported_module in 'linkgrammar$', 'clinkgrammar', '_clinkgrammar', 'lg_testutils':
@@ -786,13 +791,11 @@ class HEnglishLinkageTestCase(unittest.TestCase):
 "\nLEFT-WALL we are.v from the planet.n Gorpon[!]"
 "\n\n")
 
+
+@unittest.skipIf(NO_SQLITE_ERROR, NO_SQLITE_ERROR)
 class GSQLDictTestCase(unittest.TestCase):
     @classmethod
     def setUpClass(cls):
-        if os.name == 'nt' and \
-                -1 == clg.linkgrammar_get_configuration().lower().find('mingw'):
-            raise unittest.SkipTest("No SQL dict support yet on the MSVC build")
-
         #clg.parse_options_set_verbosity(clg.parse_options_create(), 3)
         cls.d, cls.po = Dictionary(lang='demo-sql'), ParseOptions()
 
@@ -805,10 +808,11 @@ class GSQLDictTestCase(unittest.TestCase):
 
     def test_getting_links_sat(self):
         sat_po = ParseOptions(use_sat=True)
-        if sat_po.use_sat != True:
+        if not sat_po.use_sat:
             raise unittest.SkipTest("Library not configured with SAT parser")
         linkage_testfile(self, self.d, sat_po)
 
+
 class IWordPositionTestCase(unittest.TestCase):
     @classmethod
     def setUpClass(cls):
@@ -821,12 +825,11 @@ class IWordPositionTestCase(unittest.TestCase):
     def test_en_word_positions(self):
         linkage_testfile(self, self.d_en, ParseOptions(), 'pos')
 
-# Temporarily stub out spell checking. See issue #1249 for status.
-#    def test_en_spell_word_positions(self):
-#        po = ParseOptions(spell_guess=1)
-#        if po.spell_guess == 0:
-#            raise unittest.SkipTest("Library is not configured with spell guess")
-#        linkage_testfile(self, self.d_en, po, 'pos-spell')
+    def test_en_spell_word_positions(self):
+        po = ParseOptions(spell_guess=99)
+        if po.spell_guess == 0:
+            raise unittest.SkipTest("Library is not configured with spell guess")
+        linkage_testfile(self, self.d_en, po, 'pos-spell')
 
     def test_ru_word_positions(self):
         linkage_testfile(self, Dictionary(lang='ru'), ParseOptions(), 'pos')
@@ -985,11 +988,15 @@ class XLookupListTestCase(unittest.TestCase):
     """
     @classmethod
     def setUpClass(cls):
-        cls.d_en, cls.d_sql, cls.po = Dictionary(lang='en'), Dictionary(lang='demo-sql'), ParseOptions()
+        cls.d_en, cls.po = Dictionary(lang='en'), ParseOptions()
+        if NO_SQLITE_ERROR == '':
+           cls.d_sql = Dictionary(lang='demo-sql')
 
     @classmethod
     def tearDownClass(cls):
-        del cls.d_en, cls.d_sql, cls.po
+        del cls.d_en, cls.po
+        if NO_SQLITE_ERROR == '':
+           del cls.d_sql
 
     def test_file_lookup_list_none(self):
         self.assertIsNone(clg.dictionary_lookup_list(self.d_en._obj, 'NoSuchWord'))
@@ -997,9 +1004,11 @@ class XLookupListTestCase(unittest.TestCase):
     def test_file_lookup_wild_none(self):
         self.assertIsNone(clg.dictionary_lookup_wild(self.d_en._obj, 'NoSuch*'))
 
+    @unittest.skipIf(NO_SQLITE_ERROR, NO_SQLITE_ERROR)
     def test_sql_lookup_list_none(self):
         self.assertIsNone(clg.dictionary_lookup_list(self.d_sql._obj, 'NoSuchWord'))
 
+    @unittest.skipIf(NO_SQLITE_ERROR, NO_SQLITE_ERROR)
     def test_sql_lookup_wild_none(self):
         self.assertIsNone(clg.dictionary_lookup_wild(self.d_sql._obj, 'NoSuch*'))
 
@@ -1009,6 +1018,8 @@ class XLookupListTestCase(unittest.TestCase):
 
     def test_file_lookup_list_no_subscr(self):
         dictnode = clg.dictionary_lookup_list(self.d_en._obj, 'test')
+        # The following supposes there are no idioms with the word "test".
+        # If such idioms are added, filter out the subscript "._I".
         self.assertEqual(sorted([dictnode[i].string for i in range(len(dictnode))]), [sm('test.n'), sm('test.v')])
         for i in range(len(dictnode)):
             self.assertIn("(", str(dictnode[i].exp), "Missing expression")
@@ -1017,11 +1028,13 @@ class XLookupListTestCase(unittest.TestCase):
             elif dictnode[i].string == sm('test.v'):
                 self.assertIsNone(dictnode[i].file)
 
+    @unittest.skipIf(NO_SQLITE_ERROR, NO_SQLITE_ERROR)
     @unittest.skip("FIXME: Cannot lookup with SUBSCRIPT_MARK")
     def test_sql_lookup_list_subscr(self):
         dictnode = clg.dictionary_lookup_list(self.d_en._obj, sm('test.n'))
         self.assertEqual(dictnode[0].string, sm('test.n'))
 
+    @unittest.skipIf(NO_SQLITE_ERROR, NO_SQLITE_ERROR)
     @unittest.skip("FIXME: It returns a dot subscript instead of SUBSCRIPT_MARK")
     def test_sql_lookup_list_no_subscr(self):
         dictnode = clg.dictionary_lookup_list(self.d_sql._obj, 'test')
@@ -1039,6 +1052,7 @@ class XLookupListTestCase(unittest.TestCase):
         for dn in dictnode:
             self.assertIsNotNone(re.search('test.*' + sm('.') + 'n', dn.string), 'Bad word {}'.format(dn.string))
 
+    @unittest.skipIf(NO_SQLITE_ERROR, NO_SQLITE_ERROR)
     def test_sql_lookup_wild_any_subscript(self):
         dictnode = clg.dictionary_lookup_wild(self.d_en._obj, 't*')
         self.assertTrue(len(dictnode) > 40, 'Missing words (only {} found)'.format(len(dictnode)))
@@ -1094,21 +1108,29 @@ class XExp_resolving_test(unittest.TestCase):
 
 # Currently, the dictionary creating function sets the generation mode if
 # the "test" parse-option has "generate" in its value list. So it must be
-# set so before the call to Dictionary(). When the second argument of
-# Sentence is evaluated, it initializes the test parse-option to a null
-# string.
+# set so before the call to Dictionary(). tearDownClass() takes care to
+# get out of generation mode.
 class YGenerationTestCase(unittest.TestCase):
     """
     Generation mode tests.
     """
+
+    @classmethod
+    def setUpClass(cls):
+        cls.po = ParseOptions(test='generate')
+
+    @classmethod
+    def tearDownClass(cls):
+        cls.po = ParseOptions() # Reset the "test" parse option
+        del cls.po
+
     def test_getting_linkages_file_dict(self):
-        ParseOptions(test='generate')
-        linkages = Sentence((clg.WILDCARD_WORD + ' ') * 5, Dictionary(lang='lt'), ParseOptions()).parse()
+        linkages = Sentence((clg.WILDCARD_WORD + ' ') * 5, Dictionary(lang='lt'), self.po).parse()
         self.assertTrue(len(linkages) > 0, "No linkages")
 
+    @unittest.skipIf(NO_SQLITE_ERROR, NO_SQLITE_ERROR)
     def test_getting_linkages_sql_dict(self):
-        ParseOptions(test='generate')
-        linkages = Sentence((clg.WILDCARD_WORD + ' ') * 4, Dictionary(lang='demo-sql'), ParseOptions()).parse()
+        linkages = Sentence((clg.WILDCARD_WORD + ' ') * 4, Dictionary(lang='demo-sql'), self.po).parse()
         self.assertTrue(len(linkages) > 0, "No linkages")
 
 
@@ -1229,12 +1251,85 @@ class ZRULangTestCase(unittest.TestCase):
              'облачк.=', '=а.ndnpi',
              '.', 'RIGHT-WALL'])
 
+
+class ZTHLangTestCase(unittest.TestCase):
+    def test_thai(self):
+        linkage_testfile(self, Dictionary(lang='th'), ParseOptions())
+
+
 class ZXDictDialectTestCase(unittest.TestCase):
     def test_dialect(self):
         linkage_testfile(self, Dictionary(lang='en'), ParseOptions(dialect='headline'), 'dialect')
 
+
+# Test for some catastrophic failures in displaying word expressions.
+# FIXME: This is a very small subset of the tests that are needed to
+# cover the correctness of dict_display_word_expr().
+class ZZdict_display_word_expr(unittest.TestCase):
+    @classmethod
+    def setUpClass(cls):
+       cls.d, cls.po = Dictionary(), ParseOptions()
+
+    @classmethod
+    def tearDownClass(cls):
+       del cls.d, cls.po
+
+    def test_nonexistent_word(self):
+       out = clg.dict_display_word_expr(self.d._obj, 'xxxdummy', self.po._obj)
+       self.assertIsNone(out)
+
+    def test_unsubscripted_word(self):
+       out = clg.dict_display_word_expr(self.d._obj, 'test', self.po._obj)
+       self.assertIsNotNone(out, 'Word "test" not found')
+       self.assertIn(' test.n ', out)
+
+    def test_subscripted_word(self):
+       out = clg.dict_display_word_expr(self.d._obj, 'test.v', self.po._obj)
+       self.assertIsNotNone(out, 'Word "test.v" not found')
+       self.assertIn(' test.v ', out)
+       self.assertNotIn(' test.n ', out)
+
+    def test_wildcard(self):
+       ltdict = Dictionary('lt')
+       out = clg.dict_display_word_expr(ltdict._obj, '*', self.po._obj)
+       # The output contains at least this number of non-empty lines.
+       self.assertTrue(len(list(out.splitlines())) > 1800)
+
+    def test_macros(self):
+       out = clg.dict_display_word_expr(self.d._obj, 'test/m', self.po._obj)
+       self.assertIn('<common-const-noun>:', out)
+       self.assertIn('<verb-pl,i>', out)
+
+    def test_disjuncts(self):
+       # dict_display_word_expr doesn't trigger reading the default cost_max.
+       self.po.disjunct_cost = clg.linkgrammar_get_dict_max_disjunct_cost(self.d._obj)
+
+       out = clg.dict_display_word_expr(self.d._obj, 'a//', self.po._obj)
+       self.assertIn('Token "a" disjuncts:', out)
+       #self.assertRegex(out, r'a: \[\d+] \d+\.\d+\s*=  <> Ds\*\*x\+')
+       self.assertIn(' a.eq ', out)
+
+    def test_disjunct_macros(self):
+       self.po.disjunct_cost = clg.linkgrammar_get_dict_max_disjunct_cost(self.d._obj)
+       # Here we use the word "test" that has many disjuncts, in a try to
+       # trigger memory handling bugs, if exist.
+       out = clg.dict_display_word_expr(self.d._obj, 'test//m', self.po._obj)
+       self.assertIn('Token "test" disjuncts:', out)
+       self.assertIn('<b-minus>: B*w- &', out)
+
+    def test_low_level_exp(self):
+       out = clg.dict_display_word_expr(self.d._obj, 'a/l', self.po._obj)
+       self.assertRegex(out, r'e=(0[xX])?[0-9a-fA-F]+: CONNECTOR Ds\*\*x\+ cost=0.000')
+
+
 #############################################################################
 
+# Note on the linkage order of next(linkage):
+# tests.py uses add_eqcost_linkage_order(), which sorts linkages with
+# identical metrics according to their diagram character values. Thus the
+# order of linkages with the identical metrics doesn't depend on their
+# linkage order of the LG library, which may vary across releases.
+
 def linkage_testfile(self, lgdict, popt, desc=''):
     """
     Reads sentences and their corresponding
@@ -1326,6 +1421,12 @@ def linkage_testfile(self, lgdict, popt, desc=''):
         # It ends with an empty line
         elif line[0] == 'P':
             if line[1] == '\n' and len(wordpos) > 1:
+                # Spell guesses may vary between spell packages. We assume
+                # here that those that we test here always exist (and thus know
+                # their relative order) and skip the rest.
+                if '~' in wordpos or '&' in wordpos:
+                    while getwordpos(linkage) != wordpos:
+                        linkage = next(linkages, None)
                 self.assertEqual(getwordpos(linkage), wordpos, "at {}:{}".format(testfile, lineno))
                 wordpos = None
             else:
