@@ -45,7 +45,7 @@
  */
 bool is_stem(const char* w)
 {
-	const char *subscrmark = strchr(w, SUBSCRIPT_MARK);
+	const char *subscrmark = get_word_subscript(w);
 
 	if (NULL == subscrmark) return false;
 	if (subscrmark == w) return false;
@@ -111,7 +111,11 @@ Dictionary dictionary_create_lang(const char * lang)
 	/* If an sql database exists, try to read that. */
 	if (check_db(lang))
 	{
+#if HAVE_SQLITE3
 		dictionary = dictionary_create_from_db(lang);
+#else
+		return NULL;
+#endif /* HAVE_SQLITE3 */
 	}
 
 	/* Fallback to a plain-text dictionary */
