@@ -14,7 +14,6 @@
 #include "api-types.h"
 #include "dict-api.h"
 #include "dict-common.h"
-#include "dict-defines.h" // For MAX_WORD
 #include "error.h"
 #include "idiom.h"
 #include "string-set.h"
@@ -74,9 +73,10 @@ static bool is_idiom_string(const char * s)
  * This is the same as s, but with a postfix of ".I". */
 static const char * build_idiom_word_name(Dictionary dict, const char * s)
 {
-	char buff[2*MAX_WORD];
+	size_t n = strlen(s);
+	char *buff = alloca(n+5);
 
-	size_t n = lg_strlcpy(buff, s, sizeof(buff));
+	strcpy(buff, s);
 	buff[n] = SUBSCRIPT_MARK;
 	buff[n + 1] = '_';
 	buff[n + 2] = 'I';
@@ -101,7 +101,7 @@ static Dict_node * make_idiom_Dict_nodes(Dictionary dict, const char * string)
 	Dict_node * dn = NULL;
 	char * s = strdupa(string);
 	const char * t;
-	const char *sm = strchr(s, SUBSCRIPT_MARK);
+	const char *sm = get_word_subscript(s);
 
 	for (t = s; NULL != s; t = s)
 	{
@@ -312,7 +312,7 @@ void insert_idiom(Dictionary dict, Dict_node * dn)
  */
 bool is_idiom_word(const char * s)
 {
-	const char *sm = strchr(s, SUBSCRIPT_MARK);
+	const char *sm = get_word_subscript(s);
 
 	if (NULL == sm) return false;
 	if ((sm[1] == '_') && (sm[2] == 'I') && (sm[3] == '\0')) return true;
