@@ -18,23 +18,22 @@
 #include "disjunct-utils.h"             // Disjunct
 #include "extract-links.h"
 #include "fast-match.h"
+#include "memory-pool.h"
 #include "utilities.h"                  // Windows rand_r()
 #include "linkage/linkage.h"
 #include "tokenize/word-structures.h"   // Word_Struct
 
 //#define RECOUNT
-
 //#define DEBUG_X_TABLE
-#ifdef DEBUG_X_TABLE
-#undef DEBUG_X_TABLE
-#define DEBUG_X_TABLE(...) __VA_ARGS__
-#else
-#define DEBUG_X_TABLE(...)
-#endif
 
 typedef struct Parse_choice_struct Parse_choice;
 
-/* The parse_choice is used to extract links for a given parse */
+/* Parse_choice records a parse of the word range set[0]->lw to
+ * set[1]->rw, when the middle disjunct md is of word set[0]->rw
+ * (which is always equal to set[1]->lw, link[0]->rw, and link[1]->lw).
+ * See make_choice() below.
+ * The number of linkages in this parse is the multiplication of the
+ * counts of the two Parse_set elements. */
 typedef struct Parse_set_struct Parse_set;
 struct Parse_choice_struct
 {
@@ -44,17 +43,20 @@ struct Parse_choice_struct
 	Disjunct    *md;     /* the chosen disjunct for the middle word */
 };
 
+/* Parse_set serves as a header of Parse_choice chained elements, that
+ * describe the possible parses with the specified null_count, using
+ * tracons l_id and r_id on words lw and rw, correspondingly. */
 struct Parse_set_struct
 {
-	short          lw, rw; /* left and right word index */
+	short          lw, rw;     /* left and right word index */
 	unsigned int   null_count; /* number of island words */
-	int            l_id, r_id; /* pending, unconnected connectors */
+	int            l_id, r_id; /* tracons on words lw, rw */
 
-	s64 count;      /* The number of ways to parse. */
+	count_t count;             /* The number of ways to parse. */
 #ifdef RECOUNT
-	s64 recount;  /* Exactly the same as above, but counted at a later stage. */
-	s64 cut_count;  /* Count only low-cost parses, i.e. below the cost cutoff */
-	//double cost_cutoff;
+	count_t recount;  /* Exactly the same as above, but counted at a later stage. */
+	count_t cut_count;  /* Count only low-cost parses, i.e. below the cost cutoff */
+	//float cost_cutoff;
 #undef RECOUNT
 #define RECOUNT(X) X
 #else
@@ -74,10 +76,12 @@ struct Pset_bucket_struct
 struct extractor_s
 {
 	unsigned int   x_table_size;
-	unsigned int   log2_x_table_size;
-	Pset_bucket ** x_table;  /* Hash table */
+	unsigned int   log2_x_table_size; /* Not used */
+	Pset_bucket ** x_table;           /* Hash table */
 	Parse_set *    parse_set;
 	Word           *words;
+	Pool_desc *    Pset_bucket_pool;
+	Pool_desc *    Parse_choice_pool;
 	bool           islands_ok;
 	bool           sort_match_list;
 
@@ -88,31 +92,20 @@ struct extractor_s
 /**
  * The first thing we do is we build a data structure to represent the
  * result of the entire parse search.  There will be a set of nodes
- * built for each call to the count() function that returned a non-zero
+ * built for each call to the do_count() function that returned a non-zero
  * value, AND which is part of a valid linkage.  Each of these nodes
  * represents a valid continuation, and contains pointers to two other
  * sets (one for the left continuation and one for the right
  * continuation).
  */
 
-static void free_set(Parse_set *s)
-{
-	Parse_choice *p, *xp;
-	if (s == NULL) return;
-	for (p=s->first; p != NULL; p = xp)
-	{
-		xp = p->next;
-		xfree((void *)p, sizeof(*p));
-	}
-}
-
 static Parse_choice *
 make_choice(Parse_set *lset, Connector * llc, Connector * lrc,
             Parse_set *rset, Connector * rlc, Connector * rrc,
-            Disjunct *md)
+            Disjunct *md, extractor_t* pex)
 {
 	Parse_choice *pc;
-	pc = (Parse_choice *) xalloc(sizeof(*pc));
+	pc = pool_alloc(pex->Parse_choice_pool);
 	pc->next = NULL;
 	pc->set[0] = lset;
 	pc->link[0].link_name = NULL;
@@ -151,11 +144,11 @@ static void put_choice_in_set(Parse_set *s, Parse_choice *pc)
 static void record_choice(
     Parse_set *lset, Connector * llc, Connector * lrc,
     Parse_set *rset, Connector * rlc, Connector * rrc,
-    Disjunct *md, Parse_set *s)
+    Disjunct *md, Parse_set *s, extractor_t* pex)
 {
 	put_choice_in_set(s, make_choice(lset, llc, lrc,
 	                                 rset, rlc, rrc,
-	                                 md));
+	                                 md, pex));
 }
 
 /**
@@ -165,7 +158,7 @@ static void record_choice(
  * table.  Probably should make use of the actual number of disjuncts,
  * rather than just the number of words.
  */
-extractor_t * extractor_new(int nwords, unsigned int ranstat)
+extractor_t * extractor_new(int nwords, unsigned int ranstat, bool generation)
 {
 	int log2_table_size;
 	extractor_t * pex;
@@ -175,8 +168,10 @@ extractor_t * extractor_new(int nwords, unsigned int ranstat)
 	pex->rand_state = ranstat;
 
 	/* Alloc the x_table */
-	if (nwords > 96) {
-		log2_table_size = 15 + nwords / 48;
+	if (generation) {
+		log2_table_size = 28;
+	} else if (nwords >= 72) {
+		log2_table_size = 15 + nwords / 36;
 	} else if (nwords >= 10) {
 		log2_table_size = 14 + nwords / 24;
 	} else if (nwords >= 4) {
@@ -184,17 +179,27 @@ extractor_t * extractor_new(int nwords, unsigned int ranstat)
 	} else {
 		log2_table_size = 5;
 	}
-	/* if (log2_table_size > 21) log2_table_size = 21; */
+	/* At nwords>=252, log2_table_size is 15+7=22. */
 	pex->log2_x_table_size = log2_table_size;
 	pex->x_table_size = (1 << log2_table_size);
 
-	DEBUG_X_TABLE(
+#ifdef DEBUG_X_TABLE
 		printf("Allocating x_table of size %u (nwords %d)\n",
 		       pex->x_table_size, nwords);
-	)
+#endif /* DEBUG_X_TABLE */
+
 	pex->x_table = (Pset_bucket**) xalloc(pex->x_table_size * sizeof(Pset_bucket*));
 	memset(pex->x_table, 0, pex->x_table_size * sizeof(Pset_bucket*));
 
+	pex->Pset_bucket_pool =
+		pool_new(__func__, "Pset_bucket",
+		         /*num_elements*/1024, sizeof(Pset_bucket),
+		         /*zero_out*/false, /*align*/false, /*exact*/false);
+	pex->Parse_choice_pool =
+		pool_new(__func__, "Parse_choice",
+		         /*num_elements*/1024, sizeof(Parse_choice),
+		         /*zero_out*/false, /*align*/false, /*exact*/false);
+
 	return pex;
 }
 
@@ -205,33 +210,26 @@ extractor_t * extractor_new(int nwords, unsigned int ranstat)
  */
 void free_extractor(extractor_t * pex)
 {
-	unsigned int i;
-	Pset_bucket *t, *x;
 	if (!pex) return;
 
-	DEBUG_X_TABLE(int N = 0;)
-	for (i=0; i<pex->x_table_size; i++)
+#ifdef DEBUG_X_TABLE
+	int N = 0;
+	for (unsigned int i = 0; i < pex->x_table_size; i++)
 	{
-		DEBUG_X_TABLE(int c = 0;)
-		for (t = pex->x_table[i]; t!= NULL; t=x)
-		{
-			DEBUG_X_TABLE(c++;)
-			x = t->next;
-			free_set(&t->set);
-			xfree((void *) t, sizeof(Pset_bucket));
-		}
-		DEBUG_X_TABLE(
-			if (c > 0)
-				;//printf("I %d: chain %d\n", i, c);
-			else
-				N++;
-		)
+		int c = 0;
+		for (Pset_bucket *t = pex->x_table[i]; t != NULL; t = t->next)
+			c++;
+
+		if (c > 0)
+			;//printf("I %d: chain %d\n", i, c);
+		else
+			N++;
 	}
-	DEBUG_X_TABLE(
-		printf("Used x_table %u/%u %.2f%%\n",
-				 pex->x_table_size-N, pex->x_table_size,
-				 100.0f*(pex->x_table_size-N)/pex->x_table_size);
-	)
+	printf("Used x_table %u/%u %.2f%%\n",
+	       pex->x_table_size-N, pex->x_table_size,
+	       100.0f*(pex->x_table_size-N)/pex->x_table_size);
+#endif /* DEBUG_X_TABLE */
+
 	pex->parse_set = NULL;
 
 	//printf("Freeing x_table of size %d\n", pex->x_table_size);
@@ -239,6 +237,9 @@ void free_extractor(extractor_t * pex)
 	pex->x_table_size = 0;
 	pex->x_table = NULL;
 
+	pool_delete(pex->Pset_bucket_pool);
+	pool_delete(pex->Parse_choice_pool);
+
 	xfree((void *) pex, sizeof(extractor_t));
 }
 
@@ -254,7 +255,8 @@ static Pset_bucket * x_table_pointer(int lw, int rw,
 	Pset_bucket *t;
 	int l_id = (NULL != le) ? le->tracon_id : lw;
 	int r_id = (NULL != re) ? re->tracon_id : rw;
-	t = pex->x_table[pair_hash(pex->x_table_size, lw, rw, l_id, r_id, null_count)];
+	unsigned int hash = pair_hash(lw, rw, l_id, r_id, null_count);
+	t = pex->x_table[hash & (pex->x_table_size-1)];
 
 	for (; t != NULL; t = t->next) {
 		if ((t->set.l_id == l_id) && (t->set.r_id == r_id) &&
@@ -270,10 +272,10 @@ static Pset_bucket * x_table_store(int lw, int rw,
                                   Connector *le, Connector *re,
                                   unsigned int null_count, extractor_t * pex)
 {
-	Pset_bucket *t, *n;
+	Pset_bucket **t, *n;
 	unsigned int h;
 
-	n = (Pset_bucket *) xalloc(sizeof(Pset_bucket));
+	n = pool_alloc(pex->Pset_bucket_pool);
 	n->set.lw = lw;
 	n->set.rw = rw;
 	n->set.null_count = null_count;
@@ -283,10 +285,10 @@ static Pset_bucket * x_table_store(int lw, int rw,
 	n->set.first = NULL;
 	n->set.tail = NULL;
 
-	h = pair_hash(pex->x_table_size, lw, rw, n->set.l_id, n->set.r_id, null_count);
-	t = pex->x_table[h];
-	n->next = t;
-	pex->x_table[h] = n;
+	h = pair_hash(lw, rw, n->set.l_id, n->set.r_id, null_count);
+	t = &pex->x_table[h & (pex->x_table_size -1)];
+	n->next = *t;
+	*t = n;
 	return n;
 }
 
@@ -350,11 +352,13 @@ Parse_set * mk_parse_set(fast_matcher_t *mchxt,
 {
 	int start_word, end_word, w;
 	Pset_bucket *xt;
-	Count_bin * count;
+	Count_bin *count;
 
-	assert(null_count < 0x7fff, "mk_parse_set() called with null_count < 0.");
+	if (!valid_nearest_words(le, re, lw, rw)) return NULL;
 
-	count = table_lookup(ctxt, lw, rw, le, re, null_count);
+	assert(null_count < 0x7fff, "Called with null_count < 0.");
+
+	count = table_lookup(ctxt, lw, rw, le, re, null_count, NULL);
 
 	/* If there's no counter, then there's no way to parse. */
 	if (NULL == count) return NULL;
@@ -410,7 +414,7 @@ Parse_set * mk_parse_set(fast_matcher_t *mchxt,
 					dummy = dummy_set(lw, w, null_count-1, pex);
 					record_choice(dummy, NULL, NULL,
 									  pset, dis->right, NULL,
-									  dis, &xt->set);
+									  dis, &xt->set, pex);
 					RECOUNT({xt->set.recount += pset->recount;})
 				}
 			}
@@ -422,7 +426,7 @@ Parse_set * mk_parse_set(fast_matcher_t *mchxt,
 				dummy = dummy_set(lw, w, null_count-1, pex);
 				record_choice(dummy, NULL, NULL,
 								  pset,  NULL, NULL,
-								  NULL, &xt->set);
+								  NULL, &xt->set, pex);
 				RECOUNT({xt->set.recount += pset->recount;})
 			}
 		}
@@ -459,7 +463,33 @@ Parse_set * mk_parse_set(fast_matcher_t *mchxt,
 	RECOUNT({xt->set.recount = 0;})
 	for (w = start_word; w < end_word; w++)
 	{
-		size_t mlb = form_match_list(mchxt, w, le, lw, re, rw);
+		/* Start of nonzero leftcount/rightcount range cache check. */
+		Connector *fml_re = re;       /* For form_match_list() only */
+
+		if (le != NULL)
+		{
+			if (no_count(ctxt, 0, le, w - le->nearest_word, null_count)) continue;
+			if ((re != NULL) && (re->farthest_word <= w))
+			{
+				if (no_count(ctxt, 1, re, w - re->farthest_word, null_count))
+					fml_re = NULL;
+			}
+		}
+		else
+		{
+			/* Here re != NULL. */
+			if (no_count(ctxt, 1, re, w - re->farthest_word, null_count)) continue;
+		}
+		/* End of nonzero leftcount/rightcount range cache check. */
+
+		Disjunct ***mlcl = NULL, ***mlcr = NULL;
+
+		if (le != NULL)
+			mlcl = get_cached_match_list(ctxt, 0, w, le);
+		if (re != NULL && ((le == NULL) || (re->farthest_word <= w)))
+			mlcr = get_cached_match_list(ctxt, 1, w, re);
+
+		size_t mlb = form_match_list(mchxt, w, le, lw, fml_re, rw, mlcl, mlcr);
 		if (pex->sort_match_list) sort_match_list(mchxt, mlb);
 
 		for (size_t mle = mlb; get_match_list_element(mchxt, mle) != NULL; mle++)
@@ -503,11 +533,30 @@ Parse_set * mk_parse_set(fast_matcher_t *mchxt,
 						              lw, w, le, d->left,
 						              lnull_count, pex);
 
-					ls_exists =
-						ls[0] != NULL || ls[1] != NULL || ls[2] != NULL || ls[3] != NULL;
+					if (ls[0] != NULL || ls[1] != NULL || ls[2] != NULL || ls[3] != NULL)
+					{
+						ls_exists = true;
+						/* Evaluate using the left match, but not the right */
+						Parse_set* rset = mk_parse_set(mchxt, ctxt,
+						                               w, rw, d->right, re,
+						                               rnull_count, pex);
+						if (rset != NULL)
+						{
+							for (i=0; i<4; i++)
+							{
+								if (ls[i] == NULL) continue;
+								/* this ordering is probably not consistent with
+								 * that needed to use list_links */
+								record_choice(ls[i], le, d->left,
+								              rset,  NULL /* d->right */,
+								              re,  /* the NULL indicates no link*/
+								              d, &xt->set, pex);
+								RECOUNT({xt->set.recount += (w_count_t)ls[i]->recount * rset->recount;})
+							}
+						}
+					}
 				}
 
-
 				if (Rmatch && (ls_exists || le == NULL))
 				{
 					rs[0] = mk_parse_set(mchxt, ctxt,
@@ -528,64 +577,47 @@ Parse_set * mk_parse_set(fast_matcher_t *mchxt,
 						rs[3] = mk_parse_set(mchxt, ctxt,
 						              w, rw, d->right, re,
 						              rnull_count, pex);
-				}
 
-				for (i=0; i<4; i++)
-				{
-					/* This ordering is probably not consistent with that
-					 * needed to use list_links. (??) */
-					if (ls[i] == NULL) continue;
-					for (j=0; j<4; j++)
+					if ((rs[0] != NULL || rs[1] != NULL || rs[2] != NULL || rs[3] != NULL))
 					{
-						if (rs[j] == NULL) continue;
-						record_choice(ls[i], le, d->left,
-						              rs[j], d->right, re,
-						              d, &xt->set);
-						RECOUNT({xt->set.recount += ls[i]->recount * rs[j]->recount;})
-					}
-				}
-
-				if (ls_exists)
-				{
-					/* Evaluate using the left match, but not the right */
-					Parse_set* rset = mk_parse_set(mchxt, ctxt,
-					                        w, rw, d->right, re,
-					                        rnull_count, pex);
-					if (rset != NULL)
-					{
-						for (i=0; i<4; i++)
+						if (le == NULL)
 						{
-							if (ls[i] == NULL) continue;
-							/* this ordering is probably not consistent with
-							 * that needed to use list_links */
-							record_choice(ls[i], le, d->left,
-							              rset,  NULL /* d->right */,
-							              re,  /* the NULL indicates no link*/
-							              d, &xt->set);
-							RECOUNT({xt->set.recount += ls[i]->recount * rset->recount;})
+							/* Evaluate using the right match, but not the left */
+							Parse_set* lset = mk_parse_set(mchxt, ctxt,
+							                               lw, w, le, d->left,
+							                               lnull_count, pex);
+
+							if (lset != NULL)
+							{
+								for (j=0; j<4; j++)
+								{
+									if (rs[j] == NULL) continue;
+									/* this ordering is probably not consistent with
+									 * that needed to use list_links */
+									record_choice(lset, NULL /* le */,
+									              d->left,  /* NULL indicates no link */
+									              rs[j], d->right, re,
+									              d, &xt->set, pex);
+									RECOUNT({xt->set.recount += lset->recount * rs[j]->recount;})
+								}
+							}
 						}
-					}
-				}
-				else if ((le == NULL) && (rs[0] != NULL ||
-				     rs[1] != NULL || rs[2] != NULL || rs[3] != NULL))
-				{
-					/* Evaluate using the right match, but not the left */
-					Parse_set* lset = mk_parse_set(mchxt, ctxt,
-					                        lw, w, le, d->left,
-					                        lnull_count, pex);
-
-					if (lset != NULL)
-					{
-						for (j=0; j<4; j++)
+						else
 						{
-							if (rs[j] == NULL) continue;
-							/* this ordering is probably not consistent with
-							 * that needed to use list_links */
-							record_choice(lset, NULL /* le */,
-							              d->left,  /* NULL indicates no link */
-							              rs[j], d->right, re,
-							              d, &xt->set);
-							RECOUNT({xt->set.recount += lset->recount * rs[j]->recount;})
+							for (i=0; i<4; i++)
+							{
+								/* This ordering is probably not consistent with that
+								 * needed to use list_links. (??) */
+								if (ls[i] == NULL) continue;
+								for (j=0; j<4; j++)
+								{
+									if (rs[j] == NULL) continue;
+									record_choice(ls[i], le, d->left,
+									              rs[j], d->right, re,
+									              d, &xt->set, pex);
+									RECOUNT({xt->set.recount += ls[i]->recount * rs[j]->recount;})
+								}
+							}
 						}
 					}
 				}
@@ -603,12 +635,12 @@ Parse_set * mk_parse_set(fast_matcher_t *mchxt,
 static bool set_node_overflowed(Parse_set *set)
 {
 	Parse_choice *pc;
-	s64 n = 0;
+	w_count_t n = 0;
 	if (set == NULL || set->first == NULL) return false;
 
 	for (pc = set->first; pc != NULL; pc = pc->next)
 	{
-		n  += pc->set[0]->count * pc->set[1]->count;
+		n  += (w_count_t)pc->set[0]->count * pc->set[1]->count;
 		if (PARSE_NUM_OVERFLOW < n) return true;
 	}
 	return false;
@@ -631,17 +663,16 @@ static bool set_overflowed(extractor_t * pex)
 }
 
 /**
- * This is the top level call that computes the whole parse_set.  It
- * points whole_set at the result.  It creates the necessary hash
- * table (x_table) which will be freed at the same time the
- * whole_set is freed.
+ * This is the top level call that computes the whole parse-set.
+ * It creates the necessary hash table (x_table) which will be freed at
+ * the same time the parse-set related memory is freed.
  *
  * This assumes that do_parse() has been run, and that the count_context
  * is filled with the values thus computed.  This function is structured
  * much like do_parse(), which wraps the main workhorse do_count().
  *
  * If the number of linkages gets huge, then the counts can overflow.
- * We check if this has happened when verifying the parse set.
+ * We check if this has happened when verifying the parse-set.
  * This routine returns TRUE iff an overflow occurred.
  */
 
@@ -658,7 +689,6 @@ bool build_parse_set(extractor_t* pex, Sentence sent,
 		mk_parse_set(mchxt, ctxt,
 		             -1, sent->length, NULL, NULL, null_count+1, pex);
 
-
 	return set_overflowed(pex);
 }
 
@@ -677,6 +707,8 @@ void check_link_size(Linkage lkg)
  */
 static void issue_link(Linkage lkg, bool lr, Disjunct *md, Link *link)
 {
+	if (link[0].lc == NULL) return; /* no link to generate */
+
 	if (link->rc != NULL)
 	{
 		check_link_size(lkg);
@@ -689,29 +721,71 @@ static void issue_link(Linkage lkg, bool lr, Disjunct *md, Link *link)
 
 static void issue_links_for_choice(Linkage lkg, Parse_choice *pc)
 {
-	if (pc->link[0].lc != NULL) { /* there is a link to generate */
-		issue_link(lkg, /*lr*/false, pc->md, &pc->link[0]);
-	}
-	if (pc->link[1].lc != NULL) {
-		issue_link(lkg, /*lr*/true, pc->md, &pc->link[1]);
-	}
+	issue_link(lkg, /*lr*/false, pc->md, &pc->link[0]);
+	issue_link(lkg, /*lr*/true, pc->md, &pc->link[1]);
 }
 
+/**
+ * Recursively find the \p index'th path in the parse-set tree provided by
+ * \p Parse_set and construct the corresponding linkage into the result
+ * parameter \p lkg.
+ *
+ * In order to generate all the possible linkages, the top-level function
+ * is repetitively invoked, when \p index is changing from 0 to
+ * \c num_linkages_found-1 (by extract_links(), see process_linkages()).
+ *
+ * How it works:
+ *
+ * Each "level" in the parse-set tree consists of a linked lists of
+ * Parse_choice elements. Each such element is pointed to by a
+ * Parse_choice element of an upper level. Each parse_choice element
+ * contains two Parse_set elements, that are denoted below as S0 and S1.
+ *
+ * The algo is based on our knowledge of the exact number of paths in each
+ * Parse_set element. Note that the count of the root Parse_set (used at
+ * the top-level invocation) is equal to num_linkages_found.
+ *
+ * Each list_links() invocation is done with an \p index parameter
+ * within the range of 0 to \c set->count-1 in order to extract all the
+ * paths from this set. All the \p index values in that range are used.
+ *
+ * First a selection of the Parse_choice element within the given
+ * Parse_set (with cardinality c) is done.
+ * We know that:
+ *             c-1
+ * set->count = ∑ S0ₘ * S1ₘ
+ *             m=0
+ * when S0ₘ and S1ₘ are the number of elements in S0  and S1
+ * (correspondingly) of the m'th Parse_choice element in the chain.
+ *
+ * The linkage paths of of a Parse_set element are distributed between its
+ * Parse_choice elements, each has its share ((S0ₘ * S1ₘ) for the m'th
+ * element). We scan these Parse_choice elements until we find the element
+ * that corresponds to \p index. A new index (called Nindex below) is
+ * computed, which has the property of ranging between 0 and
+ * (S0ₘ * S1ₘ)-1. It is used to further select a path in the selected
+ * Parse_choice element m. To that end we need to use its S0 and S1
+ * components in all possible combinations (when the total number of
+ * combinations is (S0ₘ * S1ₘ)):
+ *
+ * For S0: (Nindex % pc->set[0]->count) ranges from 0 to (S0ₘ-1).
+ * For S1: (Nindex / pc->set[0]->count) ranges from 0 to (S1ₘ-1).
+ */
 static void list_links(Linkage lkg, const Parse_set * set, int index)
 {
-	 Parse_choice *pc;
-	 s64 n;
+	Parse_choice *pc;
+	count_t n; /* No overflow - see extract_links() and process_linkages() */
 
-	 if (set == NULL || set->first == NULL) return;
-	 for (pc = set->first; pc != NULL; pc = pc->next) {
-		  n = pc->set[0]->count * pc->set[1]->count;
-		  if (index < n) break;
-		  index -= n;
-	 }
-	 assert(pc != NULL, "walked off the end in list_links");
-	 issue_links_for_choice(lkg, pc);
-	 list_links(lkg, pc->set[0], index % pc->set[0]->count);
-	 list_links(lkg, pc->set[1], index / pc->set[0]->count);
+	if (set == NULL || set->first == NULL) return;
+	for (pc = set->first; pc != NULL; pc = pc->next) {
+		n = pc->set[0]->count * pc->set[1]->count;
+		if (index < n) break;
+		index -= n;
+	}
+	assert(pc != NULL, "walked off the end in list_links");
+	issue_links_for_choice(lkg, pc);
+	list_links(lkg, pc->set[0], index % pc->set[0]->count);
+	list_links(lkg, pc->set[1], index / pc->set[0]->count);
 }
 
 static void list_random_links(Linkage lkg, unsigned int *rand_state,
@@ -771,3 +845,25 @@ void extract_links(extractor_t * pex, Linkage lkg)
 		list_links(lkg, pex->parse_set, index);
 	}
 }
+
+static void mark_used_disjunct(Parse_set *set, bool *disjunct_used)
+{
+	if (set == NULL || set->first == NULL) return;
+
+	for (Parse_choice *pc = set->first; pc != NULL; pc = pc->next)
+	{
+		if (pc->md->ordinal != -1)
+			disjunct_used[pc->md->ordinal] = true;
+	}
+}
+
+void mark_used_disjuncts(extractor_t *pex, bool *disjunct_used)
+{
+	assert(pex->x_table != NULL, "x_table==NULL");
+
+	for (unsigned int i = 0; i < pex->x_table_size; i++)
+	{
+		for (Pset_bucket *t = pex->x_table[i]; t != NULL; t = t->next)
+			mark_used_disjunct(&t->set, disjunct_used);
+	}
+}
