@@ -1346,7 +1346,7 @@ static void reset_last_criterion(multiset_table *cmt, const char *ctiterion)
  */
 static bool can_form_link(const char *s, const char *t, const char *e)
 {
-	if (islower(*t)) t++; /* Skip head-dependent indicator */
+	if (islower((unsigned char)*t)) t++; /* Skip head-dependent indicator */
 	while (is_connector_name_char(*s))
 	{
 		if (*s != *t) return false;
@@ -2187,7 +2187,7 @@ unsigned int pp_and_power_prune(Sentence sent, Tracon_sharing *ts,
 	if ((num_deleted > 0) && !no_mlink)
 	{
 		pc.ml = build_mlink_table(sent, ml);
-		print_time(opts, "Built_mlink_table%s", pc.ml ? "" : " (empty)");
+		print_time(opts, "Built mlink_table%s", pc.ml ? "" : " (empty)");
 		if (pc.ml != NULL)
 		{
 			if (null_count == 0)
@@ -2204,7 +2204,7 @@ unsigned int pp_and_power_prune(Sentence sent, Tracon_sharing *ts,
 		if ((num_deleted > 0) && !no_mlink)
 		{
 			pc.ml = build_mlink_table(sent, ml);
-			print_time(opts, "Built_mlink_table%s", pc.ml ? "" : " (empty)");
+			print_time(opts, "Built mlink_table%s", pc.ml ? "" : " (empty)");
 			if (pc.ml != NULL)
 			{
 				if (null_count == 0)
