@@ -25,24 +25,23 @@ typedef struct Tconnector_struct Tconnector;
 struct Tconnector_struct
 {
 	Tconnector * next;
-	const Exp *e; /* a CONNECTOR_type element from which to get the connector  */
-	int exp_pos;  /* the position in the originating expression */
+	Exp *e; /* a CONNECTOR_type element from which to get the connector  */
 };
 
 typedef struct clause_struct Clause;
 struct clause_struct
 {
 	Clause * next;
-	double cost;
-	double maxcost;
+	float cost;
+	float maxcost;
 	Tconnector * c;
 };
 
 typedef struct
 {
-	double cost_cutoff;
 	Pool_desc *Tconnector_pool;
 	Pool_desc *Clause_pool;
+	float cost_cutoff;
 	int exp_pos;
 } clause_context;
 
@@ -75,40 +74,6 @@ static void free_clause_list(Clause *c, clause_context *ct)
 }
 #endif
 
-#if 0 /* old stuff */
-/**
- * reverse the order of the list e.  destructive
- */
-static Tconnector * Treverse(Tconnector *e)
-{
-	Tconnector * head, *x;
-	head = NULL;
-	while (e != NULL) {
-		x = e->next;
-		e->next = head;
-		head = e;
-		e = x;
-	}
-	return head;
-}
-
-/**
- * reverse the order of the list e.  destructive
- */
-static Connector * reverse(Connector *e)
-{
-	Connector * head, *x;
-	head = NULL;
-	while (e != NULL) {
-		x = e->next;
-		e->next = head;
-		head = e;
-		e = x;
-	}
-	return head;
-}
-#endif
-
 /**
  * Builds a new list of connectors that is the catenation of e1 with e2.
  * does not effect lists e1 or e2.   Order is maintained.
@@ -127,16 +92,8 @@ static Tconnector * catenate(Tconnector * e1, Tconnector * e2, Pool_desc *tp)
 		preve->next = newe;
 		preve = newe;
 	}
-	for (;e2 != NULL; e2 = e2->next)
-	{
-		newe = pool_alloc(tp);
-		*newe = *e2;
 
-		preve->next = newe;
-		preve = newe;
-	}
-
-	newe->next = NULL;
+	preve->next = e2;
 	return head.next;
 }
 
@@ -148,14 +105,29 @@ static Tconnector * build_terminal(Exp *e, clause_context *ct)
 	Tconnector *c = pool_alloc(ct->Tconnector_pool);
 	c->e = e;
 	c->next = NULL;
-	c->exp_pos = ct->exp_pos++;
+	c->e->pos = ct->exp_pos++;
 	return c;
 }
 
+static void debug_last(Clause *c, Clause **c_last, const char *type)
+{
+#ifdef DEBUG
+	if ((c == NULL) || (c_last == NULL)) return;
+	do
+	{
+		if (c->next == NULL)
+		{
+			assert(c == *c_last, "%s", type);
+			return;
+		}
+	} while ((c = c->next));
+#endif
+}
+
 /**
  * Build the clause for the expression e.  Does not change e
  */
-static Clause * build_clause(Exp *e, clause_context *ct)
+static Clause * build_clause(Exp *e, clause_context *ct, Clause **c_last)
 {
 	Clause *c = NULL, *c1, *c2, *c3, *c4, *c_head;
 
@@ -169,16 +141,17 @@ static Clause * build_clause(Exp *e, clause_context *ct)
 		c1->maxcost = 0.0;
 		for (Exp *opd = e->operand_first; opd != NULL; opd = opd->operand_next)
 		{
-			c2 = build_clause(opd, ct);
+			c2 = build_clause(opd, ct, NULL);
 			c_head = NULL;
 			for (c3 = c1; c3 != NULL; c3 = c3->next)
 			{
 				for (c4 = c2; c4 != NULL; c4 = c4->next)
 				{
-					double maxcost = MAX(c3->maxcost,c4->maxcost);
+					float maxcost = MAX(c3->maxcost,c4->maxcost);
 					if (maxcost + e->cost > ct->cost_cutoff) continue;
 
 					c = pool_alloc(ct->Clause_pool);
+					if ((c_head == NULL) && (c_last != NULL)) *c_last = c;
 					c->cost = c3->cost + c4->cost;
 					c->maxcost = maxcost;
 					c->c = catenate(c3->c, c4->c, ct->Tconnector_pool);
@@ -193,31 +166,29 @@ static Clause * build_clause(Exp *e, clause_context *ct)
 			c1 = c_head;
 		}
 		c = c1;
+		if ((c != NULL) && (c->next == NULL) && (c_last != NULL)) *c_last = c;
+		debug_last(c, c_last, "AND_type");
 	}
 	else if (e->type == OR_type)
 	{
-		c = build_clause(e->operand_first, ct);
+		Clause *or_last = NULL;;
+		Clause *last;
+
+		c = build_clause(e->operand_first, ct, &or_last);
 		/* we'll catenate the lists of clauses */
 		for (Exp *opd = e->operand_first->operand_next; opd != NULL; opd = opd->operand_next)
 		{
-			c1 = build_clause(opd, ct);
+			c1 = build_clause(opd, ct, &last);
 			if (c1 == NULL) continue;
 			if (c == NULL)
-			{
 				c = c1;
-			}
 			else
-			{
-				for (c2 = c; ; c2 = c2->next)
-				{
-					if (NULL == c2->next)
-					{
-						c2->next = c1;
-						break;
-					}
-				}
-			}
+				or_last->next = c1;
+			or_last = last;
 		}
+
+		if (c_last != NULL) *c_last = (or_last == NULL) ? last : or_last;
+		debug_last(c, c_last, "OR_type");
 	}
 	else if (e->type == CONNECTOR_type)
 	{
@@ -226,6 +197,7 @@ static Clause * build_clause(Exp *e, clause_context *ct)
 		c->cost = 0.0;
 		c->maxcost = 0.0;
 		c->next = NULL;
+		if (c_last != NULL) *c_last = c;
 	}
 	else
 	{
@@ -244,8 +216,8 @@ static Clause * build_clause(Exp *e, clause_context *ct)
 		 */
 		c1->maxcost += e->cost;
 		/* Note: The above computation is used as a saving shortcut in
-		 * build_clause(). If it is changed here, it needs to be changed
-		 * there too. */
+		 * the inner loop of AND_type. If it is changed here, it needs to be
+		 * changed there too. */
 	}
 	return c;
 }
@@ -256,7 +228,7 @@ static Clause * build_clause(Exp *e, clause_context *ct)
  */
 static Disjunct *
 build_disjunct(Sentence sent, Clause * cl, const char * string,
-               const gword_set *gs, double cost_cutoff, Parse_Options opts)
+               const gword_set *gs, float cost_cutoff, Parse_Options opts)
 {
 	Disjunct *dis, *ndis;
 	Pool_desc *connector_pool = NULL;
@@ -264,11 +236,11 @@ build_disjunct(Sentence sent, Clause * cl, const char * string,
 	dis = NULL;
 	for (; cl != NULL; cl = cl->next)
 	{
-		if (NULL == cl->c) continue; /* no connectors */
+		if (unlikely(NULL == cl->c)) continue; /* no connectors */
 		if (cl->maxcost > cost_cutoff) continue;
 
 #if USE_SAT_SOLVER
-		if (NULL == sent) /* For the SAT-parser, until fixed. */
+		if (opts->use_sat_solver) /* For the SAT-parser, until fixed. */
 		{
 			ndis = xalloc(sizeof(Disjunct));
 		}
@@ -286,7 +258,7 @@ build_disjunct(Sentence sent, Clause * cl, const char * string,
 			Connector *n = connector_new(connector_pool, t->e->condesc, opts);
 			Connector **loc = ('-' == t->e->dir) ? &ndis->left : &ndis->right;
 
-			n->exp_pos = t->exp_pos;
+			n->exp_pos = t->e->pos;
 			n->multi = t->e->multi;
 			n->farthest_word = t->e->farthest_word;
 			n->next = *loc;   /* prepend the connector to the current list */
@@ -328,7 +300,7 @@ build_disjunct(Sentence sent, Clause * cl, const char * string,
 }
 
 Disjunct *build_disjuncts_for_exp(Sentence sent, Exp* exp, const char *word,
-                                  const gword_set *gs, double cost_cutoff,
+                                  const gword_set *gs, float cost_cutoff,
                                   Parse_Options opts)
 {
 	Clause *c;
@@ -336,20 +308,31 @@ Disjunct *build_disjuncts_for_exp(Sentence sent, Exp* exp, const char *word,
 	clause_context ct = { 0 };
 
 	ct.cost_cutoff = cost_cutoff;
-	ct.Clause_pool = pool_new(__func__, "Clause",
-	                   /*num_elements*/4096, sizeof(Clause),
-	                   /*zero_out*/false, /*align*/false, /*exact*/false);
-	ct.Tconnector_pool = pool_new(__func__, "Tconnector",
-	                   /*num_elements*/32768, sizeof(Tconnector),
-	                   /*zero_out*/false, /*align*/false, /*exact*/false);
+	if (sent->Clause_pool == NULL)
+	{
+		ct.Clause_pool = pool_new(__func__, "Clause",
+		                          /*num_elements*/4096, sizeof(Clause),
+		                          /*zero_out*/false, /*align*/false, /*exact*/false);
+		ct.Tconnector_pool = pool_new(__func__, "Tconnector",
+		                              /*num_elements*/32768, sizeof(Tconnector),
+		                              /*zero_out*/false, /*align*/false, /*exact*/false);
+		/* Keep for freeing at the caller. */
+		sent->Clause_pool = ct.Clause_pool;
+		sent->Tconnector_pool = ct.Tconnector_pool;
+	}
+	else
+	{
+		ct.Clause_pool = sent->Clause_pool;
+		ct.Tconnector_pool = sent->Tconnector_pool;
+	}
 
-	// printf("%s\n", exp_stringify(exp));
-	c = build_clause(exp, &ct);
+	// printf("%s\n", lg_exp_stringify(exp));
+	c = build_clause(exp, &ct, NULL);
 	// print_clause_list(c);
 	dis = build_disjunct(sent, c, word, gs, cost_cutoff, opts);
 	// print_disjunct_list(dis);
-	pool_delete(ct.Tconnector_pool);
-	pool_delete(ct.Clause_pool);
+	pool_reuse(ct.Clause_pool);
+	pool_reuse(ct.Tconnector_pool);
 	return dis;
 }
 
