@@ -759,11 +759,11 @@ bool strtodC(const char *s, float *r)
 {
 	char *err;
 
-#if defined(HAVE_LOCALE_T) && !defined(__sun__)
-	double val = strtod_l(s, &err, get_C_LC_NUMERIC());
+#if defined(HAVE_LOCALE_T) && !defined(__sun__) && !defined(__OpenBSD__)
+	float val = strtof_l(s, &err, get_C_LC_NUMERIC());
 #else
 	/* dictionary_setup_locale() invokes setlocale(LC_NUMERIC, "C") */
-	double val = strtod(s, &err);
+	float val = strtof(s, &err);
 #endif /* HAVE_LOCALE_T */
 
 	if ('\0' != *err) return false; /* *r unaffected */
